/*************************************************************************************************************************

 Program        : SHOW_MEMORY.EXE
 Author         : Nick de Smith         NICK@NCDLAB.ULCC.AC.UK
 Creation date  : 23-Aug-90

  1990 Nick de Smith and Applied Telematics Group Ltd.
 This software is supplied for information only. No guarantee is supplied for this software, and no liability
 will be accepted for any action resulting from the use of this software or the information contained herein.
 Under no circumstances may this software be used for commercial gain, including its sale, lease or loan.
 This software may be copied only with the inclusion of this copyright notice.
 The author is prepared to enter into correspondance with interested parties, but will not necessarily
 maintain this software. Having said all that, enjoy it!

 Edit   Edit date       By      Why
  02    25-Sep-90       NMdS    Add in percentage rounding fiddle factor.
  01    23-Aug-90       NMdS    New

 Show system memory usage. Requires CMEXEC privilege.
 This code, I believe, is definitive. That is to say that it implements the same algorithms as used by SHOW MEMORY (as
 of VMS V5.3-1), which I presume to be correct. Hopefully my implementation of these algorithms is correct, and so this
 will kill another one of the perenial Info-VAX woolly arguments.

 Build with:    $ CC SHOW_MEMORY
                $ LINK SHOW_MEMORY,SYS$SYSTEM:SYS.STB/SELECT
                $ RUN SHOW_MEMORY

*************************************************************************************************************************/

#module SHOW_MEMORY     "V01.02"

#include        descrip                                                 /* Define DSC$xxx - VAX descriptors  */
#include        psldef                                                  /* Define PSL$xxx - Access mode symbols  */
#include        ssdef                                                   /* Define SS$xxxx - System status codes  */
#include        stsdef                                                  /* Define STS$M_xxxx - VMS status bits  */

/* Useful macro to check returned status of a system routine */
#define ss_check( command ) {                   \
        long ss_status = (command);             \
        if ( (ss_status & STS$M_SUCCESS) == 0 ) { \
                return ss_status;               \
        }                                       \
}

/*************************************************************************************************************************

                                        S H O W _ M E M O R Y

 Show the memory usage by this VMS system.
 All displayed percentages are with respect to the ENABLED memory, not the total on the system.

*************************************************************************************************************************/

SHOW_MEMORY()
{
        static $DESCRIPTOR( x_format_1, "There are !UL physical page!%S (!UL.!AD Mb) on this system" );
        static $DESCRIPTOR( x_format_10,"!_all of which is enabled" );
        static $DESCRIPTOR( x_format_11,"!_of which only !UL page!%S (!UL.!AD Mb) are enabled" );
        static $DESCRIPTOR( x_format_2, "Pages Inuse: !UL (!UL%)  Free: !UL (!UL%)  Modified: !UL (!UL%)  Bad: !UL" );
        static $DESCRIPTOR( x_format_3, "Of the pages in use, !UL (!UL%) are permanently allocated to VMS" );
        static struct dsc$descriptor_d x_output = {
                0, DSC$K_DTYPE_T, DSC$K_CLASS_D, 0 };
        unsigned long l_physical_pages;                                 /* Physical pages in enabled for use  */
        unsigned long l_free_pages;                                     /* Pages on free list    */
        unsigned long l_modified_pages;                                 /* Pages on modified list   */
        unsigned long l_vms_pages;                                      /* Pages in use by VMS    */
        unsigned long l_bad_pages;                                      /* Pages marked "bad"    */
        unsigned long l_real_physical_pages;                            /* Physical pages connected to system  */
        unsigned long args[] = { 6,
                &l_physical_pages, &l_free_pages, &l_modified_pages, &l_vms_pages, &l_bad_pages, &l_real_physical_pages };
        static int SIZE_MEMORY();                                       /* CMEXEC routine to get memory size.  */

        ss_check( SYS$CMEXEC( SIZE_MEMORY, args ) )                     /* Call the data collector   */

        ss_check( LIB$SYS_FAO(
                &x_format_1             ,                               /* Format control string   */
                0                       ,                               /* No resultant length    */
                &x_output               ,                               /* Output string    */
                l_real_physical_pages   ,                               /* Number of physical pages   */
                l_real_physical_pages >> 11,                            /* Number of megabytes of physical memory */
                2, &"00255075"[ ((l_real_physical_pages >> 9) & 3) * 2 ] ) ) /* Fractional part of Mb   */
        ss_check( LIB$PUT_OUTPUT( &x_output ) )

        if ( l_real_physical_pages == l_physical_pages ) {              /* If all memory is enabled...   */
                ss_check( LIB$SYS_FAO(                                  /* ...say all memory is enabled   */
                        &x_format_10            ,                       /* Format control string   */
                        0                       ,                       /* No resultant length    */
                        &x_output               ) )                     /* Output string    */
        } else {
                ss_check( LIB$SYS_FAO(                                  /* Say how much memory is enabled  */
                        &x_format_11            ,                       /* Format control string   */
                        0                       ,                       /* No resultant length    */
                        &x_output               ,                       /* Output string    */
                        l_physical_pages        ,                       /* Number of enabled physical pages  */
                        l_physical_pages >> 11  ,                       /* Enabled megabytes of physical memory  */
                        2, &"00255075"[ ((l_physical_pages >> 9) & 3) * 2 ] ) ) /* Fractional part of Mb  */
        }
        ss_check( LIB$PUT_OUTPUT( &x_output ) )

#define DIVISOR         l_physical_pages                                /* percentages are WRT "physical_pages"  */
#define FIDDLE_FACTOR   (DIVISOR / 2)                                   /* Fiddle factor to round up/down the %tages */
#define PERCENTAGE( value ) (((value) * 100 + FIDDLE_FACTOR) / DIVISOR) /* Return percentage of "value" WRT "physical_pages" */

        ss_check( LIB$SYS_FAO(
                &x_format_2     ,                                       /* Format control string   */
                0               ,                                       /* No resultant length    */
                &x_output       ,                                       /* Output string    */
                l_physical_pages - l_free_pages - l_modified_pages,     /* Inuse page count    */
                PERCENTAGE( l_physical_pages - l_free_pages - l_modified_pages ),
                l_free_pages    ,                                       /* Free page count    */
                PERCENTAGE( l_free_pages ),
                l_modified_pages,                                       /* Modified page count    */
                PERCENTAGE( l_modified_pages ),
                l_bad_pages     ) )                                     /* Count of bad pages    */
        ss_check( LIB$PUT_OUTPUT( &x_output ) )

        ss_check( LIB$SYS_FAO(
                &x_format_3     ,                                       /* Format control string   */
                0               ,                                       /* No resultant length    */
                &x_output       ,                                       /* Output string    */
                l_vms_pages     ,                                       /* Pages in use by VMS    */
                PERCENTAGE( l_vms_pages ) ) )
        ss_check( LIB$PUT_OUTPUT( &x_output ) )

        return SS$_NORMAL;
}

/*************************************************************************************************************************

                                        S I Z E _ M E M O R Y

 Calculate the amount of memory in use by this VMS system.

  1990 Nick de Smith and Applied Telematics Group Ltd.
 This software is supplied for information only. No guarantee is supplied for this software, and no liability
 will be accepted for any action resulting from the use of this software or the information contained herein.
 Under no circumstances may this software be used for commercial gain, including its sale, lease or loan.
 This software may be copied only with the inclusion of this copyright notice.
 The author is prepared to enter into correspondance with interested parties, but will not necessarily
 maintain this software. Having said all that, enjoy it!

 This code, I believe, is definitive. That is to say that it implements the same algorithms as used by SHOW MEMORY,
 which I presume to be correct. Hopefully my implementation of these algorithms is correct, and so this will kill
 another one of the perenial Info-VAX woolly arguments 8-).

 This is a "stand-alone" routine that can be called from any VMS language. It must be called at EXEC mode or via a
 SYS$CMEXEC call with an argument list.
 The code in the routine is effectively a C representation of the DCL SHOW MEMORY command. I acknowledge the use of
 the VMS V5.3-1 fiche in determining the algorithms to use.
 No scan is done of the bad page list as we do not need to differentiate between soft and hard memory errors - we just
 take the total bad count.
 A interesting feature of this routine is that it returns the "real_physical_pages" (the number of pages of memory
 physically in the machine) as well as "physical_pages" (the number of pages that are enabled). This latter number will
 always be less than or equal to the former. The reason for this is that the SYSGEN parameter PHYSICALPAGES can be used
 to artificially constrain a system to use less memory than there really is, in order to test systems with differing
 memory sizes without having to put in or haul out physical boards. As the routine returns both values, it is possible
 to check a) how much "real physical" memory there is actually on the system, and b) how much of it VMS is allowed
 to use. The DCL command SHOW MEMORY only displays the enabled memory.
 This routine must be linked with SYS$SYSTEM:SYS.STB/Selective_Search.

 Edit   Edit date       By              Why
  01    20-Sep-90       Nick de Smith   First attempt - with thanks to the SHOW MEMORY code.

*************************************************************************************************************************/

globalref unsigned long SCH$GL_FREECNT;                                 /* Number of free pages    */
#define PFN$C_BADPAGLST 0x2                                             /* ...Offset to count of bad pages in longwords */
globalref unsigned long SCH$GL_MFYCNT;                                  /* Number of modified pages   */
globalref unsigned long PFN$GL_PHYPGCNT;                                /* Physical pages for process use  */
globalref unsigned long MMG$GL_PHYPGCNT;                                /* Configured (SYSGEN) limit on pages  */
globalref unsigned char *EXE$GL_RPB;                                    /* => Restart Parameter Block   */

/* From $NDTDEF  - Nexus Device Type codes      */
#define NDT$_MPM0       0x0040                                          /* Multiport Memory 0    */
#define NDT$_MPM3       0x0043                                          /*     ''      ''   3    */

/* From $RPBDEF - Restart Parameter Block       */
#define RPB$L_MEMDSC    0xBC                                            /* Offset into RPB for memory descriptors */
#define RPB$C_NMEMDSC   0x8                                             /* Max. number of memory descriptor.  */
typedef struct MEMDSC {
        unsigned page_count     : 24;                           /* Number of pages in this descriptor */
        unsigned tr_number      : 8;                            /* Adaptor type number   */
        unsigned base_pfn       : 32;                           /* Base PFN for this memory  */
} MEMDSC, *MEMDSC_PTR;

#define RPB$L_BOOTR5    0x30                                            /* Offset into RPB copy of boot time R5  */
#define RPB$M_MPM       0x00000800                                      /* Use MultiPortMemory only   */
#define RPB$M_USEMPM    0x00001000                                      /* Use MultiPortMemory as well as local memory */

#define RPB$L_BADPGS    0x104                                           /* Count of bad pages at boot time  */

unsigned long
SIZE_MEMORY(
        unsigned long * al_physical_pages       ,                       /* => Returned physical page count  */
        unsigned long * al_free_pages           ,                       /* =>    ''    free      ''   ''  */
        unsigned long * al_modified_pages       ,                       /* =>    ''    modified  ''   ''  */
        unsigned long * al_vms_pages            ,                       /* Count of pages used by VMS   */
        unsigned long * al_bad_pages            ,                       /* Count of bad pages    */
        unsigned long * al_real_physical_pages  )                       /* "real" count of physical pages  */
{
        int             l_local_pages, l_multiport_pages;
        unsigned long   l_nmemdsc = RPB$C_NMEMDSC;                      /* Maximum number of memory descriptors  */
        MEMDSC_PTR      ar_memdsc = EXE$GL_RPB + RPB$L_MEMDSC;          /* => first memory descriptor in RPB  */
        unsigned long   l_boot_r5 = *((unsigned long *)(EXE$GL_RPB + RPB$L_BOOTR5)); /* Copy of system boot R5 value */

        *al_free_pages          = SCH$GL_FREECNT;                       /* Return number of pages on free list  */
        *al_modified_pages      = SCH$GL_MFYCNT;                        /*   ''     ''   ''   ''  '' modified list */

        l_local_pages = l_multiport_pages = 0;

        for ( ; ar_memdsc->page_count && l_nmemdsc-- ; ar_memdsc++ ) {  /* Scan each memory descriptor   */

                /* If adapter is within MPM range, then its multiport memory. An assumption is made here that all memory
                   with adaptor numbers in the range NDT$_MPM0 to NDT$_MPM3 is multiport memory, and that no multiport
                   memory has an adaptor number outside of this range.                */

                if ( ar_memdsc->tr_number >= NDT$_MPM0 && ar_memdsc->tr_number <= NDT$_MPM3 ) {
                        l_multiport_pages += ar_memdsc->page_count;     /* Its a multiport memory descriptor  */
                } else {
                        l_local_pages += ar_memdsc->page_count;         /* Its local memory only   */
                }
        }

        /* Check the boot flags (/R5:xxxxxxxx) for the memory configuration (really for 11/782s only)   */

        if ( l_boot_r5 & RPB$M_MPM ) {                                  /* If this is a multiprocessor configuration... */
                *al_real_physical_pages = l_multiport_pages;            /* ...then we are only using shared memory */
        } else {
                *al_real_physical_pages = l_local_pages;                /* The system is using the local memory. */
                if ( l_boot_r5 & RPB$M_USEMPM ) {                       /* If USEMPM is set, we are using MPM as well... */
                        *al_real_physical_pages += l_multiport_pages;   /* ...so add in the multiport memory pages */
                }
        }

        /* Check to see if the SYSGEN PHYSICALPAGES parameter has been used to limit the usable number of pages  */

        if ( *al_real_physical_pages <= MMG$GL_PHYPGCNT ) {             /* If the physical memory count is lower... */
                *al_physical_pages = *al_real_physical_pages;           /* ...than the maximum allowed by SYSGEN, use it */
        } else {
                *al_physical_pages = MMG$GL_PHYPGCNT;                   /* Usable pages have been limited by SYSGEN */
        }
        *al_vms_pages = *al_physical_pages - PFN$GL_PHYPGCNT;           /* Get pages used by VMS   */

        *al_bad_pages = *((unsigned long *)(EXE$GL_RPB + RPB$L_BADPGS)); /* # of bad pages at boot   */
        *al_bad_pages += *((unsigned long *)(&SCH$GL_FREECNT + PFN$C_BADPAGLST)); /* # of bad pages since boot  */

        *al_vms_pages -= *al_bad_pages;                                 /* Don't count bad pages to VMS   */

        return SS$_NORMAL;
}
